module Views.ForceDirectedGraph exposing (..)

import Browser
import Browser.Events
import Color
import Config.ForceDirectedGraph as Local
import Data.DataLoader exposing (Entry)
import Force
import Graph exposing (Edge, Graph, NodeContext, NodeId)
import Html exposing (Html)
import Html.Attributes exposing (style)
import Html.Events exposing (onClick, onInput)
import Html.Events.Extra.Mouse as Mouse
import Html.Lazy exposing (lazy2)
import Json.Decode as Decode
import List.Extra
import List.Statistics
import Model exposing (Global, Model)
import Msg exposing (Msg)
import Process exposing (sleep)
import Task
import TypedSvg exposing (circle, g, line, svg, text_, title)
import TypedSvg.Attributes exposing (alignmentBaseline, class, fill, shapeRendering, stroke, textAnchor, textRendering, viewBox)
import TypedSvg.Attributes.InPx exposing (cx, cy, fontSize, r, strokeWidth, x, x1, x2, y, y1, y2)
import TypedSvg.Core exposing (Attribute, Svg, text)
import TypedSvg.Types exposing (AlignmentBaseline(..), AnchorAlignment(..), Paint(..), ShapeRendering(..), TextRendering(..), px)
import Utils exposing (flag)


w : Float
w =
    1000


h : Float
h =
    500


view : Global -> Local.Model -> Html Msg
view global local =
    Html.div [ style "width" "100%" ]
        [ Html.div [ style "display" "flex", style "flex-direction" "row", style "align-items" "center" ]
            [ Html.div []
                [ Html.h2 [ style "margin-top" "0" ]
                    [ Html.text "Force directed Graph" ]
                , Html.button [ Local.ChangeForces 0 |> propagate |> onClick ]
                    [ Html.i [ Html.Attributes.class "fa-solid fa-repeat fa-xl" ] [] ]
                , Html.button [ Local.ChangeForces 1 |> propagate |> onClick ]
                    [ Html.i [ Html.Attributes.class "fa-solid fa-repeat fa-xl" ] [] ]
                , Html.button [ Local.ChangeForces 2 |> propagate |> onClick ]
                    [ Html.i [ Html.Attributes.class "fa-solid fa-repeat fa-xl" ] [] ]
                ]
            , Html.button [ Local.RecalcStart |> propagate |> onClick, style "margin-left" "auto" ]
                [ Html.i [ Html.Attributes.class "fa-solid fa-repeat fa-xl" ] [] ]
            , Html.div [ style "margin-left" "auto" ]
                [ Html.text "Jahr: "
                , allYears
                    global.data
                    |> List.map
                        (\c ->
                            Html.option []
                                [ if c > 0 then
                                    Html.text <| String.fromInt c

                                  else
                                    Html.text "Alle Jahre"
                                ]
                        )
                    |> Html.select [ onInput (\s -> propagate <| Local.ChangeYear <| Maybe.withDefault 0 <| String.toInt s) ]
                ]
            ]
        , Html.div [ style "display" "flex", style "flex-direction" "row" ]
            [ if local.loading then
                Html.div [ style "flex-grow" "1" ] [ Html.text "Loading..." ]

              else
                Html.div [ style "flex-grow" "1", style "position" "relative" ]
                    [ svg
                        [ viewBox 0 0 w h, style "position" "absolute", style "width" "100%", style "height" "100%" ]
                        [ Graph.edges local.graph
                            |> List.map (linkElement local.graph local.highlight)
                            |> g [ class [ "links" ] ]
                        , Graph.nodes local.graph
                            |> List.map (nodeElement local global.data)
                            |> g [ class [ "nodes" ] ]
                        ]
                    ]
            , Html.div [ Html.Attributes.class "legend", style "width" "19%" ]
                [ Html.div [ style "display" "flex" ]
                    [ Html.h3 []
                        [ Html.text "Legende"
                        ]
                    , if not <| local.highlight == "" then
                        Html.button [ Local.Highlight "" |> propagate |> onClick, style "margin-left" "auto" ]
                            [ Html.i [ Html.Attributes.class "fa-solid fa-repeat fa-xl" ] [] ]

                      else
                        Html.div [] []
                    ]
                , List.map
                    (\ent ->
                        [ flag ent.state
                            ++ " "
                            ++ Utils.codeToName ent.state
                            |> Html.text
                        ]
                            |> Html.button [ Local.Highlight (String.toUpper ent.state) |> propagate |> onClick, style "padding-bottom" "0.2rem", style "padding-top" "0.2rem" ]
                    )
                    global.data
                    |> List.Extra.unique
                    |> Html.div []
                ]
            ]
        ]


filterForYears : List Entry -> Int -> List Entry
filterForYears initial year =
    if year > 0 then
        initial
            |> List.filter (\e -> e.year == year)

    else
        initial


allYears : List Entry -> List Int
allYears data =
    List.map .year data
        |> List.Extra.unique
        |> List.sort
        |> List.reverse
        |> List.append [ 0 ]


allSchengen : List Entry -> List String
allSchengen data =
    List.map .state data
        |> List.Extra.unique


initializeNode : NodeContext String () -> NodeContext Local.Entity ()
initializeNode ctx =
    { node = { label = Force.entity ctx.node.id ctx.node.label, id = ctx.node.id }
    , incoming = ctx.incoming
    , outgoing = ctx.outgoing
    }


firstInit : Local.Model
firstInit =
    Local.Model Graph.empty (Force.simulation []) 0 True "" 0


makeGraph : List Entry -> Int -> Graph String ()
makeGraph data year =
    let
        filteredData =
            filterForYears data year

        countryNames =
            List.map (\ent -> String.toUpper ent.state) filteredData

        consulateCountries =
            List.map (\ent -> String.toUpper ent.consulateCountry) filteredData

        allCountries =
            List.Extra.unique <| List.concat [ countryNames, consulateCountries ]

        findPos : String -> Int
        findPos s =
            Maybe.withDefault -1 <| List.Extra.findIndex (\it -> it == s) allCountries

        edges =
            List.Extra.unique <|
                List.map
                    (\ent ->
                        ( findPos <| String.toUpper ent.state
                        , findPos <| String.toUpper ent.consulateCountry
                        )
                    )
                    filteredData
    in
    Graph.fromNodeLabelsAndEdgePairs allCountries edges


postInit : List Entry -> Int -> Task.Task Never (Result Local.TaskError Local.Model)
postInit data year =
    let
        filteredData =
            filterForYears data year

        countryNames =
            List.map (\ent -> String.toUpper ent.state) filteredData

        consulateCountries =
            List.map (\ent -> String.toUpper ent.consulateCountry) filteredData

        allCountries =
            List.Extra.unique <| List.concat [ countryNames, consulateCountries ]

        findPos : String -> Int
        findPos s =
            Maybe.withDefault -1 <| List.Extra.findIndex (\it -> it == s) allCountries

        findName : Int -> String
        findName i =
            Maybe.withDefault "???" <| List.Extra.getAt i allCountries

        rowsWithBoth : String -> String -> List Entry
        rowsWithBoth s o =
            List.filter (\ent -> String.toUpper ent.state == String.toUpper s && String.toUpper ent.consulateCountry == String.toUpper o) filteredData

        graph =
            Graph.mapContexts initializeNode <| makeGraph filteredData year

        forces =
            [ Graph.nodes graph
                |> List.map
                    (\n ->
                        let
                            cc =
                                List.Extra.getAt (findName n.id |> findPos) allCountries
                                    |> Maybe.withDefault "??"

                            rows =
                                List.filter (\e -> e.state == cc) filteredData

                            acceptedRatio =
                                (List.map (\e -> Maybe.withDefault 0 <| e.totalVisasApplied) rows
                                    |> List.Statistics.mean
                                    |> Maybe.withDefault 1
                                )
                                    / (List.map (\e -> Maybe.withDefault 0 <| e.totalVisasIssued) rows
                                        |> List.Statistics.mean
                                        |> Maybe.withDefault 1
                                      )
                        in
                        ( n.id, 0.5 * acceptedRatio )
                    )
                |> Force.customManyBody 0.99

            --Force.customLinks 1
            --  (List.map link (Graph.edges graph))
            --  Force.manyBodyStrength -10 <|
            --    List.map .id <|
            --        Graph.nodes graph
            , Graph.nodes graph
                |> List.map
                    (\e ->
                        ( e.id
                        , if List.member (findName e.id) countryNames then
                            30

                          else
                            10
                        )
                    )
                |> Force.customCollision { iterations = 1, strength = 1.0 }

            --Graph.edges graph
            --  |> List.map
            --      (\e ->
            --          let
            --              meanNotIssRat =
            --                  rowsWithBoth (findName e.from) (findName e.to)
            --                      |> List.map
            --                          (\ent ->
            --                              let
            --                                  totApp =
            --                                      Maybe.withDefault 0 <| ent.totalVisasApplied
            --                                  totIss =
            --                                      Maybe.withDefault 0 <| ent.totalVisasIssued
            --                                  notIss =
            --                                      totApp - totIss
            --                                  notIssRat =
            --                                      notIss / totApp
            --                              in
            --                              if not (isNaN notIssRat) && not (notIssRat < 0) then
            --                                  notIssRat
            --                              else
            --                                  0
            --                          )
            --                      |> List.Statistics.mean
            --                      |> Maybe.withDefault 0
            --          in
            --          { source = e.from, target = e.to, distance = min w h * (1 - meanNotIssRat) / 1.5, strength = Just 1 }
            --      )
            --  |> Force.customLinks 3
            --, Graph.edges graph
            --    |> List.map
            --        (\i ->
            --            if
            --                isSchengen
            --                    (case Graph.get i.from graph of
            --                        Just ctx ->
            --                            ctx.node _ -> Graph.Node 0 (Force.entity 0 "")
            --                    )
            --                    data
            --                    && isSchengen
            --                        (case Graph.get i.to graph of
            --                            Just ctx ->
            --                                ctx.node
            --                            _ ->
            --                                Graph.Node 0 (Force.entity 0 "")
            --                        )
            --                        data
            --                    || not
            --                        (isSchengen
            --                            (case Graph.get i.from graph of
            --                                Just ctx ->
            --                                    ctx.node
            --                                _ ->
            --                                    Graph.Node 0 (Force.entity 0 "")
            --                            )
            --                            data
            --                        )
            --                    && not
            --                        (isSchengen
            --                            (case Graph.get i.to graph of
            --                                Just ctx ->
            --                                    ctx.node
            --                                _ ->
            --                                    Graph.Node 0 (Force.entity 0 "")
            --                            )
            --                            data
            --                        )
            --            then
            --                { source = i.from, target = i.to, distance = 3, strength = Just 2 }
            --            else
            --                { source = i.from, target = i.to, distance = 10, strength = Just 1 }
            --        )
            --    |> Force.customLinks 1
            , Graph.nodes graph
                |> List.map
                    (\i ->
                        if isSchengen i data then
                            ( i.id, { strength = 4.0, x = w / 2, y = h / 2, radius = min h w / 3.0 } )

                        else
                            ( i.id, { strength = 4.0, x = w / 2, y = h / 2, radius = min h w / 1.2 } )
                    )
                |> Force.customRadial
            , Force.center
                (w / 2)
                (h / 2)
            ]

        list =
            Force.computeSimulation (Force.simulation forces) <| List.map .label <| Graph.nodes graph
    in
    Task.succeed (Ok { firstInit | loading = True })
        |> Task.andThen
            (\_ ->
                Task.succeed
                    (Ok
                        (Local.Model (updateGraphWithList graph list) (Force.simulation forces) 0 False "" 0)
                    )
            )


isSchengen : Graph.Node Local.Entity -> List Entry -> Bool
isSchengen ent data =
    let
        index =
            List.map (\en -> String.toUpper en.state) data
                |> List.Extra.findIndex (\le -> le == String.toUpper ent.label.value)
                |> Maybe.withDefault -1
    in
    if index == -1 then
        False

    else
        True


updateNode : ( Float, Float ) -> NodeContext Local.Entity () -> NodeContext Local.Entity ()
updateNode ( x, y ) nodeCtx =
    let
        nodeValue =
            nodeCtx.node.label
    in
    updateContextWithValue nodeCtx { nodeValue | x = x, y = y }


updateContextWithValue : NodeContext Local.Entity () -> Local.Entity -> NodeContext Local.Entity ()
updateContextWithValue nodeCtx value =
    let
        node =
            nodeCtx.node
    in
    { nodeCtx | node = { node | label = value } }


updateGraphWithList : Graph Local.Entity () -> List Local.Entity -> Graph Local.Entity ()
updateGraphWithList =
    let
        graphUpdater value =
            Maybe.map (\ctx -> updateContextWithValue ctx value)
    in
    List.foldr (\node graph -> Graph.update node.id (graphUpdater node) graph)


calculateModel : List Entry -> Int -> Cmd Local.Msg
calculateModel data year =
    Task.perform Local.RecalcEnd (Process.sleep 0 |> Task.andThen (\_ -> postInit data year))


update : Local.Msg -> Global -> Local.Model -> ( Global, Local.Model, Cmd Msg )
update msg glob { graph, simulation, tickCount, loading, highlight, year } =
    case msg of
        Local.RecalcStart ->
            ( glob, Local.Model graph simulation tickCount True highlight year, calculateModel glob.data year |> Cmd.map propagate )

        Local.Highlight hl ->
            ( glob, Local.Model graph simulation tickCount loading hl year, Cmd.none )

        Local.ChangeYear y ->
            ( glob, Local.Model graph simulation tickCount True highlight year, calculateModel glob.data y |> Cmd.map propagate )

        Local.ChangeForces f ->
            ( glob, Local.Model graph simulation tickCount loading highlight year, Cmd.none )

        Local.SetLoading b ->
            ( glob, Local.Model graph simulation tickCount b highlight year, Cmd.none )

        Local.RecalcEnd res ->
            case res of
                Ok newModel ->
                    ( glob, { newModel | loading = False }, Cmd.none )

                Err _ ->
                    ( glob, Local.Model graph simulation tickCount False highlight year, Cmd.none )


propagate : Local.Msg -> Msg
propagate msg =
    Msg.GotForceDirectedGraph msg


subscriptions : Global -> Local.Model -> Sub Msg
subscriptions global model =
    Sub.map propagate Sub.none


linkElement : Graph Local.Entity () -> String -> Edge () -> Svg msg
linkElement graph hl edge =
    let
        source =
            Maybe.withDefault (Force.entity 0 "") <| Maybe.map (.node >> .label) <| Graph.get edge.from graph

        target =
            Maybe.withDefault (Force.entity 0 "") <| Maybe.map (.node >> .label) <| Graph.get edge.to graph
    in
    line
        [ strokeWidth 1
        , stroke <| Paint <| Color.rgba 0.5 0.5 0.5 0.3
        , x1 source.x
        , y1 source.y
        , x2 target.x
        , y2 target.y
        , class
            [ source.value
            , target.value
            , if hl == source.value || hl == target.value then
                "highlight"

              else
                ""
            ]
        ]
        []


nodeElement : Local.Model -> List Entry -> { a | id : NodeId, label : { b | x : Float, y : Float, value : String } } -> Svg Msg
nodeElement local dat node =
    let
        filteredData =
            filterForYears dat local.year

        countryNames =
            List.map (\ent -> String.toUpper ent.state) filteredData

        consulateCountries =
            List.map (\ent -> String.toUpper ent.consulateCountry) filteredData

        allCountries =
            List.Extra.unique <| List.concat [ countryNames, consulateCountries ]

        findPos : String -> Int
        findPos s =
            Maybe.withDefault -1 <| List.Extra.findIndex (\it -> it == s) allCountries

        findName : Int -> String
        findName i =
            Maybe.withDefault "???" <| List.Extra.getAt i allCountries

        cc =
            List.Extra.getAt (findName node.id |> findPos) allCountries
                |> Maybe.withDefault "??"

        rows =
            List.filter (\e -> e.state == cc) filteredData

        visasApplied =
            rows
                |> List.map (\e -> Maybe.withDefault 1 <| e.totalVisasApplied)
                |> List.sum

        baseFontSize =
            15 + visasApplied / 1000000

        highlightFontSize =
            if local.highlight == node.label.value then
                2 * baseFontSize

            else
                baseFontSize
    in
    g []
        [ text_
            [ x <| node.label.x - 8.0
            , y node.label.y
            , fontSize highlightFontSize
            , alignmentBaseline AlignmentMiddle
            , textAnchor AnchorMiddle
            , class
                [ node.label.value
                ]
            ]
            [ title [] [ text <| Utils.codeToName node.label.value ], flag node.label.value |> text ]
        ]
